#!/usr/bin/env python

# psycopg2 test suite
#
# Copyright (C) 2007-2011 Federico Di Gregorio  <fog@debian.org>
#
# psycopg2 is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In addition, as a special exception, the copyright holders give
# permission to link this program with the OpenSSL library (or with
# modified versions of OpenSSL that use the same license as OpenSSL),
# and distribute linked combinations including the two.
#
# You must obey the GNU Lesser General Public License in all respects for
# all of the code used other than OpenSSL.
#
# psycopg2 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.

from __future__ import print_function

import sys
from psycopg2cffi.tests.psycopg2_tests.testconfig import dsn
from psycopg2cffi.tests.psycopg2_tests.testutils import unittest

from psycopg2cffi.tests.psycopg2_tests import (
        test_async,
        test_bugX000,
        test_bug_gc,
        test_bug_inf_fetch_loop,
        test_cancel,
        test_connection,
        test_copy,
        test_cursor,
        test_dates,
        test_extras_dictcursor,
        test_green,
        test_lobject,
        test_module,
        test_notify,
        test_psycopg2_dbapi20,
        test_quote,
        test_transaction,
        test_types_basic,
        test_types_extras,
        )

if sys.version_info[:2] >= (2, 5):
    from psycopg2cffi.tests.psycopg2_tests import test_with
else:
    test_with = None

def test_suite():
    # If connection to test db fails, bail out early.
    import psycopg2cffi
    try:
        cnn = psycopg2cffi.connect(dsn)
    except Exception as e:
        print("Failed connection to test db:", e.__class__.__name__, e)
        print("Please set env vars 'PSYCOPG2_TESTDB*' to valid values.")
        sys.exit(1)
    else:
        cnn.close()

    suite = unittest.TestSuite()
    suite.addTest(test_async.test_suite())
    suite.addTest(test_bugX000.test_suite())
    suite.addTest(test_bug_gc.test_suite())
    suite.addTest(test_cancel.test_suite())
    suite.addTest(test_connection.test_suite())
    suite.addTest(test_copy.test_suite())
    suite.addTest(test_cursor.test_suite())
    suite.addTest(test_dates.test_suite())
    suite.addTest(test_extras_dictcursor.test_suite())
    suite.addTest(test_green.test_suite())
    suite.addTest(test_lobject.test_suite())
    suite.addTest(test_module.test_suite())
    suite.addTest(test_notify.test_suite())
    suite.addTest(test_psycopg2_dbapi20.test_suite())
    suite.addTest(test_quote.test_suite())
    suite.addTest(test_transaction.test_suite())
    suite.addTest(test_types_basic.test_suite())
    suite.addTest(test_types_extras.test_suite())
    if test_with:
        suite.addTest(test_with.test_suite())
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')
